import numpy as np
from scipy.optimize import curve_fit
from sklearn.metrics import r2_score as r2


# Functions
def powerlaw(x, a, b):
    return a * (x ** b)


def poly(x, a, b, c):
    return a * (x ** b) + c * x


def hertzian(ind, diameter, e_star):
    return (4 / 3) * e_star * ((0.5 * diameter) ** 0.5) * (ind ** 1.5)


def exponential(t, y_0, tau, y_f):
    return (y_0 - y_f) * np.exp(-(t-t[0]) * tau) + y_f


def biexponential(t, y_0, c, tau1, tau2, y_f):
    return (y_0 - y_f) * (c * np.exp(-(t-t[0]) * tau1) + (1 - c) * np.exp(-(t-t[0]) * tau2)) + y_f


def poroelastic(t, y_0, c, tau1, tau2, y_f):
    return (y_0 - y_f) * (c * np.exp(-np.sqrt(t-t[0]) * tau1) + (1 - c) * np.exp(-(t-t[0]) * tau2)) + y_f


# Fitting
def lin_fit(x, y):
    popt = np.polyfit(x, y, 1)
    y_pred = np.poly1d(popt)(x)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


def powerlaw_fit(x, y):
    x_min = min(x)
    x_max = max(x)
    y_min = min(y)
    y_max = max(y)
    x = (x - x_min) / x_max
    y = (y - y_min) / y_max
    bounds = ([0, 0], [np.inf, np.inf])
    popt, _ = curve_fit(powerlaw, x, y, bounds=bounds)
    y_pred = powerlaw(x, *popt)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred * y_max + y_min


def poly_fit(x, y):
    x_min = min(x)
    x_max = max(x)
    y_min = min(y)
    y_max = max(y)
    x = (x - x_min) / x_max
    y = (y - y_min) / y_max
    bounds = ([0, 0,  0], [np.inf, np.inf, np.inf])
    popt, _ = curve_fit(poly, x, y, bounds=bounds)
    y_pred = poly(x, *popt)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred * y_max + y_min


def hertzian_fit(x, y, diameter):
    def wrapper(ind, e_star):
        return hertzian(ind, diameter, e_star)

    x_min = min(x)
    y_min = min(y)
    x = x - x_min
    y = y - y_min

    popt, _ = curve_fit(wrapper, x, y)

    y_pred = hertzian(x, diameter, *popt)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred + y_min


def exponential_fit(x, y, bound=False):
    if bound:
        def bnd_wrapper(t, tau):
            return exponential(t, y[0], tau, y[-1])

        bounds = ([0], [np.inf])
        popt, _ = curve_fit(bnd_wrapper, x, y, bounds=bounds)
        args = [y[0], *popt, y[-1]]
    else:
        def ubnd_wrapper(t, tau, y_f):
            return exponential(t, y[0], tau, y_f)

        bounds = ([0, -np.inf], [np.inf, np.inf])
        popt, _ = curve_fit(ubnd_wrapper, x, y, bounds=bounds)
        args = [y[0], *popt]

    y_pred = exponential(x, *args)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


def biexponential_fit(x, y, bound=False):
    if bound:
        def bnd_wrapper(t, c, tau1, tau2):
            return biexponential(t, y[0], c, tau1, tau2, y[-1])

        bounds = ([0, 0, 0], [1, np.inf, np.inf])
        p0 = [0.4, 1, 0.1]
        popt, _ = curve_fit(bnd_wrapper, x, y,
                            bounds=bounds, p0=p0, jac='3-point')
        args = [y[0], *popt, y[-1]]
    else:
        def ubnd_wrapper(t, c, tau1, tau2, y_f):
            return biexponential(t, y[0], c, tau1, tau2, y_f)

        p0 = [0.4, 1, 0.1, 0]
        bounds = ([0, 0, 0, -np.inf], [1, np.inf, np.inf, np.inf])
        popt, _ = curve_fit(ubnd_wrapper, x, y,
                            bounds=bounds, p0=p0, jac='3-point')
        args = [y[0], *popt]

    y_pred = biexponential(x, *args)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


def poroelastic_fit(x, y, bound=False):
    if bound:
        def bnd_wrapper(t, c, tau1, tau2):
            return poroelastic(t, y[0], c, tau1, tau2, y[-1])

        bounds = ([0, 0, 0], [1, np.inf, np.inf])
        p0 = [0.4, 1, 0.1]
        popt, _ = curve_fit(bnd_wrapper, x, y,
                            bounds=bounds, p0=p0, jac='3-point')
        args = [y[0], *popt, y[-1]]
    else:
        def ubnd_wrapper(t, c, tau1, tau2, y_f):
            return poroelastic(t, y[0], c, tau1, tau2, y_f)

        p0 = [0.4, 1, 0.1, 0]
        bounds = ([0, 0, 0, -np.inf], [1, np.inf, np.inf, np.inf])
        popt, _ = curve_fit(ubnd_wrapper, x, y,
                            bounds=bounds, p0=p0, jac='3-point')
        args = [y[0], *popt]

    y_pred = poroelastic(x, *args)
    r2_score = r2(y, y_pred)
    return popt, r2_score, y_pred


if __name__ == '__main__':
    pass



from dash import dcc

import plotly.graph_objects as go


def render(id: str, title: str, xaxis: str) -> dcc.Graph:
    fig = make_fig(title, xaxis)

    if 'contact' in id:
        fig.add_vline(x=0, line=dict(color='red', width=1.2))
        fig.add_hline(y=0, line=dict(color='red', width=1.2))

    return dcc.Graph(
        figure=fig,
        id=id,
        mathjax=True,
        style={
            'width': '50%',
        }
    )


def make_fig(title: str, xaxis: str) -> go.Figure:
    fig = go.Figure()

    fig.update_layout(
        title={
            'text': title,
            'x': 0.5,
            'xanchor': 'center'
        },
        plot_bgcolor='white',
        paper_bgcolor='white',
        margin=dict(t=50, b=50, l=50, r=50),
        showlegend=False,
        transition={'duration': 500},
    )

    fig.update_annotations(yshift=10)

    fig.update_xaxes(
        showgrid=False,
        ticks='outside',
        tickwidth=2,
        showline=True,
        linewidth=2,
        linecolor='black',
        mirror=True,
        title_text=xaxis,
        tickprefix=r"$",
        ticksuffix=r"$",
    )

    fig.update_yaxes(
        showgrid=False,
        ticks='outside',
        tickwidth=2,
        showline=True,
        linewidth=2,
        linecolor='black',
        mirror=True,
        title_text=r"$Force \text{ (N)}$",
        tickprefix=r"$",
        ticksuffix=r"$",
    )

    return fig


def update_fig(fig: go.Figure, x, y, mode: str, color: str, hover: bool = False, y2: bool = False) -> go.Figure:
    hover_texts = None
    if hover:
        hover_texts = [f'Index: {i}' for i in range(len(x))]

    if not fig['data']:
        if mode == 'lines':
            color_arg = {'line': {'color': color}}
        elif mode == 'markers':
            color_arg = {'marker': {'color': color}}
        else:
            color_arg = {}

        trace = go.Scattergl(
            x=x,
            y=y,
            mode=mode,
            text=hover_texts,
            hoverinfo='text',
            hoverlabel={'bgcolor': 'red'},
            **color_arg,
        )

        fig.add_trace(trace)
    else:
        fig.data[0].x = x
        fig.data[0].y = y
        fig.data[0]['text'] = hover_texts

    return fig


def update_contact_line(cp: int, fig: dict | go.Figure) -> go.Figure:
    if isinstance(fig, dict):
        fig = go.Figure(fig)
    data_x = fig.data[0].x
    data_y = fig.data[0].y
    fig.layout['shapes'][0]['x0'] = data_x[cp]
    fig.layout['shapes'][0]['x1'] = data_x[cp]
    fig.layout['shapes'][1]['y0'] = data_y[cp]
    fig.layout['shapes'][1]['y1'] = data_y[cp]
    fig.layout['title']['text'] = fr"$\text{{Selected Contact Point: {cp}}}$"

    return fig



from igor2 import binarywave as bw
import numpy as np
import pandas as pd
import re
import os

from . import curves


def _get_notes(wave: dict) -> dict:
    '''Utility function for processing the 'notes' section of a .ibw file,
    the end user should not call this function'''

    note_raw = wave['wave']['note']
    # Asylum seems to store the degree sign in a broken way that
    # python can't parse, replace all occurances of this invalid
    # byte sequence with 'deg'
    note_raw = note_raw.replace(b'\xb0', b'deg')
    all_notes = note_raw.split(b'\r')
    note_dict = dict()
    for line in all_notes:
        split_line = line.split(b':')
        key = split_line[0]
        value = b':'.join(split_line[1:]).strip()
        note_dict[key.decode()] = value.decode()
    return note_dict


def _get_data(wave: dict) -> pd.DataFrame:
    '''Utility function for processing the 'data' section of a .ibw file,
    the end user should not call this function'''

    wave_labels = wave['wave']['labels']
    wave_data = wave['wave']['wData']
    labels = [label.decode() for label in wave_labels[1] if label]
    col_indices = {
        'rawz': labels.index('Raw'),
        'defl': labels.index('Defl'),
        'z': labels.index('ZSnsr')
    }

    df = dict(
        rawz=wave_data[:, col_indices['rawz']],
        z=wave_data[:, col_indices['z']],
        defl=wave_data[:, col_indices['defl']]
    )

    wave_frame = pd.DataFrame(df)
    wave_frame.loc[:, 'ind'] = wave_frame['z'] - wave_frame['defl']
    return wave_frame


def load_ibw(filename: str) -> curves.Curve:
    wave = bw.load(filename)
    data = _get_data(wave)
    notes = _get_notes(wave)
    trigger_index = int(np.argmax(data.loc[:, 'defl']))

    sample_time = wave['wave']['wave_header']['sfA'][0]
    t = np.arange(data.shape[0]) * sample_time
    data.loc[:, 't'] = t

    if notes.get('DwellTime'):
        dwell_time = float(notes['DwellTime'])
    else:
        print("Missing DwellTime")
        dwell_time = 0.0

    dwell_start_time = data['t'].loc[trigger_index]
    dwell_end_time = dwell_start_time + dwell_time

    dwell_end_index = int(np.argmin(np.abs(data.loc[:, 't'] - dwell_end_time)))
    dwell_range = [trigger_index, dwell_end_index]

    if notes.get('SpringConstant'):
        k = float(notes['SpringConstant'])
    else:
        print("Missing SpringConstant")
        k = 0.0

    data.loc[:, 'f'] = data.loc[:, 'defl'] * k

    if notes.get('InvOLS'):
        invOLS = float(notes['InvOLS'])
    else:
        print("Missing InvOLS")
        invOLS = 0.0

    this_curve = curves.Curve(
        filename=filename.split(os.path.sep)[-1],
        data=data,
        notes=notes,
        invOLS=invOLS,
        k=k,
        dwell_range=dwell_range
    )

    return this_curve


def load_curveset_ibw(folder: str, ident_labels: list[str]) -> curves.CurveSet:
    """
    Loads a set of .ibw files from a specified directory into a curves.CurveSet object.

    This function scans a directory for .ibw files associated with Asylum Atomic Force Microscopy (AFM) measurements. 
    It filters and loads these files into a CurveSet object based on specified identification labels found in the filenames. 
    These labels help distinguish between different measurements within the set.

    Parameters:
    - folder (str): The path to the directory containing .ibw files to be loaded. 
      The directory should exclusively contain relevant .ibw files created by an Asylum AFM.

    - ident_labels (list[str]): A list of substrings that are consistently present in the filenames of interest, 
      in the order they appear. These labels are used to filter and differentiate between measurements. 
      Intervening strings not included in this list will be considered as part of the unique identifiers for each measurement.

    Example:
    Consider a directory at '~/experiment' with the following files:
    - Sample1Measurement0.ibw
    - Sample1Measurement1.ibw
    - Sample2Measurement0.ibw
    - Sample2Measurement1.ibw

    To load these files into a CurveSet, use:
    >>> load_curveset_ibw('~/experiment', ['Sample', 'Measurement'])

    Returns:
    - curves.CurveSet: An object containing all loaded force curves that match the identification labels,
      organized as specified by the 'folder' and 'ident_labels' parameters.
    """

    regex_str = ""
    for label in ident_labels:
        regex_str = regex_str + label + f'(?P<{label}>.*)'
    regex_str = regex_str + '\\.ibw'
    regex = re.compile(regex_str)

    all_filenames = os.listdir(folder)
    all_matches = [regex.match(a) for a in all_filenames]

    curve_dict = dict()
    for m in all_matches:
        if not m:
            continue
        idents = tuple([m.group(label) for label in ident_labels])
        filename = m.group(0)
        filepath = os.path.join(folder, filename)

        print(f"   > Reading {filename}")
        curve_dict[idents] = load_ibw(filepath)

    curveset = curves.CurveSet(
        ident_labels=ident_labels, curve_dict=curve_dict)
    print("   > Experiment Loaded.")
    return curveset



from plotly import graph_objs as go

import PyPDF2 as pdf
import pickle
import pandas as pd
import numpy as np
import json
import ast
import io
from tqdm import tqdm
from typing import Any, Iterable
from time import sleep

from .fit import lin_fit, poly_fit, hertzian_fit, biexponential_fit
from .src.components.fig import make_fig


def read_json_file(file: str) -> dict[tuple, bool | int]:
    with open(file, 'rt') as cf:
        anno_str_dict = json.load(cf)

    anno_tuple_dict = {}
    for key, value in anno_str_dict.items():
        tuple_key = ast.literal_eval(key)
        anno_tuple_dict[tuple_key] = value

    return anno_tuple_dict


class Curve:

    _data: pd.DataFrame
    corrected: bool = False
    contact_index: int = 0
    contact_values: pd.Series | None = None
    max_ind: float
    max_f: float
    vel_ind: float
    vel_z: float
    figs_data: dict[str, tuple[np.ndarray, np.ndarray]]
    fits_data: dict[str, tuple[np.ndarray, np.ndarray]]
    indent_param: list[float]
    indent_R2: float
    hertzian_param: list[float]
    hertzian_R2: float
    dwell_param: list[float]
    dwell_R2: float
    contact_fig: go.Figure
    dwell_fig: go.Figure

    def __init__(
            self,
            filename: str,
            data: pd.DataFrame,
            notes: dict,
            invOLS: float,
            k: float,
            dwell_range: list[int],
    ) -> None:

        self.filename = filename
        self.data = data
        self.notes = notes
        self.invOLS = invOLS
        self.k = k
        self.dwell_range = dwell_range

        self.backup_data()

    def backup_data(self) -> None:
        self._data = self.data.copy()

    def restore_data(self) -> None:
        self.data = self._data.copy()

    def reduce_data(self) -> None:
        self.data.drop(['rawz', 'defl'], axis=1, inplace=True)

    def set_contact_index(self, cp: int) -> None:
        self.contact_index = cp

    def check_contact(self) -> None:
        if self.contact_index == 0:
            raise Exception('Contact index has not been set.')

    def get_contact_values(self) -> pd.Series:
        if self.contact_values is None:
            self.contact_values = self.data.iloc[self.contact_index].copy()
        return self.contact_values

    def adjust_to_contact(self) -> None:
        self.data -= self.get_contact_values()

    def get_data_between(self, first: int, last: int) -> pd.DataFrame:
        return self.data.iloc[first:last].reset_index(drop=True)

    def get_approach(self) -> pd.DataFrame:
        return self.get_data_between(0, self.dwell_range[0])

    def get_preapproach(self) -> pd.DataFrame:
        return self.get_data_between(0, self.contact_index)

    def get_indent(self) -> pd.DataFrame:
        return self.get_data_between(self.contact_index, self.dwell_range[0])

    def get_indent_until(self, ind: float) -> pd.DataFrame:
        indent = self.get_indent()
        ind_index = (np.abs(indent['ind'].to_numpy() - ind)).argmin()
        return indent.iloc[:ind_index].reset_index(drop=True)

    def get_indent_between(self, first: float, last: float) -> pd.DataFrame:
        indent = self.get_indent()
        f = indent['f'].to_numpy()
        first_index = np.argmin(np.abs(f - f[-1] * first))
        last_index = np.argmin(np.abs(f - f[-1] * last))
        return indent.iloc[first_index:last_index].reset_index(drop=True)

    def get_dwell(self) -> pd.DataFrame:
        return self.get_data_between(self.dwell_range[0], self.dwell_range[1])

    def get_repulsion(self) -> pd.DataFrame:
        dwell = self.get_dwell()
        max_ind = dwell['ind'].argmax()
        return dwell.iloc[max_ind:].reset_index(drop=True)

    def get_retract(self) -> pd.DataFrame:
        return self.get_data_between(self.dwell_range[1], -1)

    def get_trigger_data(self) -> pd.Series:
        indent = self.get_indent()
        return indent.iloc[-1].reset_index(drop=True)

    def get_approach_rates(self) -> None:
        indent = self.get_indent()
        t, ind, z = indent['t'], indent['ind'], indent['z']
        self.vel_ind = lin_fit(t, ind)[0][0]
        self.vel_z = lin_fit(t, z)[0][0]

    def correct_virtual_defl(self) -> None:
        self.check_contact()
        if self.corrected:
            self.restore_data()

        preapproach = self.get_preapproach()
        x = preapproach['z'].to_numpy()
        y = preapproach['f'].to_numpy()
        popt, _, _ = lin_fit(x, y)
        f_line = np.poly1d(popt)(self.data['z'])
        self.data['f'] = self.data['f'] - f_line

    def detect_contact(self) -> int:
        current_contact_index = self.contact_index
        approach = self.get_approach()
        max_index = len(approach) - 1
        index = max_index
        while True:
            self.set_contact_index(index)
            self.correct_virtual_defl()
            approach_new = self.get_approach()
            force = approach_new['f'].to_numpy()
            mean = np.mean(force[:index+1])

            ratio = force[index] / force[-1]
            step_size = max(1, int((0.01 * max_index) ** ratio))
            index = index - step_size
            if force[index] < mean:
                contact_index = index + 1
                break

        self.set_contact_index(current_contact_index)
        self.restore_data()
        return contact_index

    def fit_indent(self) -> tuple:
        self.check_contact()
        indent = self.get_indent()
        return poly_fit(indent['ind'], indent['f'])

    def fit_indent_until(self, probe_diameter: float, ind: float) -> tuple:
        self.check_contact()
        indent_until = self.get_indent_until(ind)
        return hertzian_fit(indent_until['ind'], indent_until['f'], probe_diameter)

    def fit_indent_between(self, probe_diameter: float, interval: list[float] = [0.0, 1.0]) -> tuple:
        self.check_contact()
        indent_between = self.get_indent_between(interval[0], interval[1])
        return hertzian_fit(indent_between['ind'], indent_between['f'], probe_diameter)

    def fit_dwell(self) -> tuple:
        dwell = self.get_dwell()
        return biexponential_fit(dwell['t'], dwell['f'])

    def get_figs_data(self, vd: bool = False, adjust: bool = False) -> None:
        if vd or adjust:
            self.restore_data()
        if vd:
            self.correct_virtual_defl()
        if adjust:
            self.adjust_to_contact()

        approach = self.get_approach()
        dwell = self.get_dwell()

        self.figs_data = {
            'approach': (approach['ind'].to_numpy(), approach['f'].to_numpy()),
            'dwell': (dwell['t'].to_numpy(), dwell['f'].to_numpy())
        }

    def get_contact_fig_plot(self) -> go.Figure:
        fig = make_fig(
            title=fr"$\text{{Selected Contact Point: {self.contact_index}}}$",
            xaxis=r"$Indentation \text{ (m)}$"
        )

        x, y = self.figs_data['approach']
        hover_texts = [f'Index: {i}' for i in range(len(x))]
        trace = go.Scattergl(
            x=x,
            y=y,
            name='Approach',
            mode='markers',
            text=hover_texts,
            hoverinfo='text',
            hoverlabel={'bgcolor': 'red'},
            marker={'color': 'blue'},
        )

        fig.add_trace(trace)

        x_line = x[self.contact_index]
        y_line = y[self.contact_index]
        fig.add_vline(x=x_line, line=dict(color='red', width=1.2))
        fig.add_hline(y=y_line, line=dict(color='red', width=1.2))
        self.contact_fig = fig
        return fig

    def get_dwell_fig_plot(self) -> go.Figure:
        fig = make_fig(
            title=r"$\text{Dwell}$",
            xaxis=r"$Time \text{ (s)}$"
        )

        x, y = self.figs_data['dwell']
        trace = go.Scattergl(
            x=x,
            y=y,
            name='Dwell',
            mode='markers',
            marker={'color': 'red'},
        )

        fig.add_trace(trace)

        self.dwell_fig = fig
        return fig

    def get_fits_data(self, probe_diameter: float, ind: float | list[float]) -> None:
        indent = self.get_indent()
        indent_x_pred = indent['ind'].to_numpy()
        self.indent_param, self.indent_R2, indent_y_pred = self.fit_indent()

        self.max_ind = max(indent['ind'])
        self.max_f = max(indent['f'])

        if isinstance(ind, float):
            self.hertzian_param, self.hertzian_R2, hertzian_y_pred = self.fit_indent_until(
                probe_diameter, ind)
            hertzian_x_pred = self.get_indent_until(ind)['ind'].to_numpy()
        elif isinstance(ind, list) and len(ind) == 2:
            self.hertzian_param, self.hertzian_R2, hertzian_y_pred = self.fit_indent_between(
                probe_diameter, interval=ind)
            hertzian_x_pred = self.get_indent_between(ind[0], ind[1])[
                'ind'].to_numpy()

        dwell = self.get_dwell()
        dwell_x_pred = dwell['t'].to_numpy()
        self.dwell_param, self.dwell_R2, dwell_y_pred = self.fit_dwell()

        self.fits_data = {
            'indent': (indent_x_pred, indent_y_pred),
            'hertzian': (hertzian_x_pred, hertzian_y_pred),
            'dwell': (dwell_x_pred, dwell_y_pred),

        }

    def add_contact_fit(self) -> go.Figure:
        x, y = self.fits_data['indent']
        trace = go.Scattergl(
            x=x,
            y=y,
            name='PolyFit',
            mode='lines',
            line={'color': 'red'},
        )
        self.contact_fig.add_trace(trace)

        x, y = self.fits_data['hertzian']
        trace = go.Scattergl(
            x=x,
            y=y,
            name='HertzianFit',
            mode='lines',
            line={
                'color': 'green',
                'width': 3,
            },
        )
        self.contact_fig.add_trace(trace)

        return self.contact_fig

    def add_dwell_fit(self) -> go.Figure:
        x, y = self.fits_data['dwell']

        trace = go.Scattergl(
            x=x,
            y=y,
            name='DwellFit',
            mode='lines',
            line={
                'color': 'green',
                'width': 3,
            },
        )
        self.dwell_fig.add_trace(trace)

        return self.dwell_fig

    def get_contact_fig_pdf(self) -> go.Figure:
        title = fr"$\text{{Selected Contact Point: {self.contact_index}}}$"
        xaxis = r"$Indentation \text{ (m)}$"
        fig = make_fig(title, xaxis)

        x, y = self.figs_data['approach']
        trace = go.Scatter(x=x, y=y, name='Approach', marker={'color': 'blue'})
        fig.add_trace(trace)

        x_line = x[self.contact_index]
        y_line = y[self.contact_index]
        fig.add_vline(x=x_line, line=dict(color='red', width=1.2))
        fig.add_hline(y=y_line, line=dict(color='red', width=1.2))

        if hasattr(self, 'fits_data'):
            x, y = self.fits_data['indent']
            trace = go.Scatter(x=x, y=y, name='PolyFit',
                               mode='lines', line={'color': 'red'})
            fig.add_trace(trace)

            x, y = self.fits_data['hertzian']
            trace = go.Scatter(x=x, y=y, name='HertzianFit',
                               mode='lines', line={'color': 'green', 'width': 3})
            fig.add_trace(trace)

        fig.update_layout(width=480, height=400)
        sleep(1)
        return fig

    def get_dwell_fig_pdf(self) -> go.Figure:
        title = r"$\text{Dwell}$"
        xaxis = r"$Time \text{ (s)}$"
        fig = make_fig(title, xaxis)

        x, y = self.figs_data['dwell']
        trace = go.Scatter(x=x, y=y, name='Dwell',
                           marker_color='red')
        fig.add_trace(trace)

        if hasattr(self, 'fits_data'):
            x, y = self.fits_data['dwell']
            trace = go.Scatter(x=x, y=y, name='DwellFit',
                               mode='lines', line={'color': 'green', 'width': 3})
            fig.add_trace(trace)

        fig.update_layout(width=480, height=400)
        sleep(1)
        return fig

    def get_all_fits(self) -> dict:
        fit_results_dict = {
            'max_ind': self.max_ind,
            'max_f': self.max_f,

            'vel_ind': self.vel_ind,
            'vel_z': self.vel_z,

            'indent_a': self.indent_param[0],
            'indent_b': self.indent_param[1],
            'indent_c': self.indent_param[2],
            'indent_R2': self.indent_R2,

            'Hertzian_E': self.hertzian_param[0],
            'Hertzian_R2': self.hertzian_R2,

            'dwell_c': self.dwell_param[0],
            'dwell_tau1': self.dwell_param[1],
            'dwell_tau2': self.dwell_param[2],
            'dwell_tauMax': max(self.dwell_param[1], self.dwell_param[2]),
            'dwell_tauMin': min(self.dwell_param[1], self.dwell_param[2]),
            'dwell_yf': self.dwell_param[3],
            'dwell_R2': self.dwell_R2,
        }

        return fit_results_dict


class CurveSet:
    vd_dict: dict[tuple[str], bool] = {}
    cp_dict: dict[tuple[str], int] = {}

    def __init__(self, ident_labels: list[str], curve_dict: dict[tuple[str], Curve]) -> None:

        self.ident_labels = ident_labels
        self.curve_dict = curve_dict
        self._index = 0

    def __iter__(self):
        self._index = 0
        return self

    def __next__(self) -> Curve:
        if self._index < len(self.curve_dict):
            key = list(self.curve_dict.keys())[self._index]
            self._index += 1
            return self.curve_dict[key]
        else:
            raise StopIteration

    def __getitem__(self, key: tuple[str]) -> Curve:
        return self.curve_dict[key]

    def items(self) -> Iterable[tuple[tuple[str], Curve]]:
        return self.curve_dict.items()

    def pickle(self, filename: str) -> None:
        with open(filename, 'wb') as f:
            pickle.dump(self, f)

    def keys(self) -> list[tuple[str]]:
        return list(self.curve_dict.keys())

    def reduce_data(self) -> None:
        for curve in self:
            curve.reduce_data()

    def remove_curve(self, key: tuple[str]) -> None:
        del self.curve_dict[key]

    def remove_unannotated(self) -> None:
        keys_to_remove = []
        for key, curve in self.items():
            if curve.contact_index == 0:
                keys_to_remove.append(key)

        for key in keys_to_remove:
            self.remove_curve(key)

    def correct_all_virtual_defl(self) -> None:
        for curve in self:
            curve.correct_virtual_defl()

    def set_vd_by_annotations(self) -> None:
        for key, value in self.vd_dict.items():
            if value:
                self[key].correct_virtual_defl()

    def set_cp_by_annotations(self) -> None:
        for key, value in self.cp_dict.items():
            self[key].set_contact_index(value)

    def update_annotations(self, annotations_dict: dict[tuple[str], Any]) -> None:
        value = next(iter(annotations_dict.values()))
        if isinstance(value, bool):
            self.vd_dict = annotations_dict
        elif isinstance(value, int):
            self.cp_dict = annotations_dict

    def update_annotations_from_file(self, file: str) -> None:
        anno_tuple_dict = read_json_file(file)
        self.update_annotations(anno_tuple_dict)

    def adjust_to_contact(self) -> None:
        for curve in self:
            curve.adjust_to_contact()

    def prepare_fit_all(self) -> None:
        self.set_cp_by_annotations()
        self.remove_unannotated()
        self.set_vd_by_annotations()
        self.adjust_to_contact()

    def get_fit_all(self, probe_diameter: float, ind: float | list[float]) -> pd.DataFrame:
        self.prepare_fit_all()
        all_fit = []

        for key, curve in tqdm(self.items()):
            fit_results_dict = {label: key_value for label,
                                key_value in zip(self.ident_labels, key)}
            curve.get_fits_data(probe_diameter, ind)
            curve.get_approach_rates()
            fit_results_dict.update(curve.get_all_fits())
            fit_results = pd.DataFrame([fit_results_dict])
            all_fit.append(fit_results)

        combined_results = pd.concat(all_fit, ignore_index=True)

        return combined_results

    def export_figures(self) -> pdf.PdfMerger:
        merger = pdf.PdfMerger()
        for key, curve in tqdm(self.items()):
            curve.get_figs_data()
            title = ""
            for label, ident in zip(self.ident_labels, key):
                title += label + ident

            title = fr"$\text{{{title}}}$"
            contact_fit_fig = curve.get_contact_fig_pdf()
            contact_fit_fig.update_layout(title={'text': title})
            contact_fit_fig_pdf = io.BytesIO(
                contact_fit_fig.to_image(format='pdf'))
            merger.append(contact_fit_fig_pdf)

            dwell_fit_fig = curve.get_dwell_fig_pdf()
            dwell_fit_fig.update_layout(title={'text': title})
            dwell_fit_fig_pdf = io.BytesIO(
                dwell_fit_fig.to_image(format='pdf'))
            merger.append(dwell_fit_fig_pdf)

        return merger



from dash import Dash, dcc, html, callback_context
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State

from typing import Protocol

from . import ids
from ..utils.utils import load


class AFM(Protocol):
    curve_keys: list[tuple[str, ...]]


def render(app: Dash) -> html.Div:
    @app.callback(
        [Output(ids.CURVE_DROPDOWN, 'options'),
         Output(ids.CURVE_DROPDOWN, 'value', allow_duplicate=True),
         Output(ids.LOG, 'children', allow_duplicate=True)],
        [Input(ids.EXPERIMENT, 'data')],
        [State(ids.EXPERIMENT_LABELS, 'value'),
         State(ids.LOG, 'children')],
        prevent_initial_call=True
    )
    def update_curve_dropdown(encoded_experiment, labels, exp_output):
        experiment: AFM = load(encoded_experiment)
        keys = experiment.curve_keys
        label_list = [label.strip() for label in labels.split(';')]

        options = []
        for idx, key in enumerate(keys):
            name = ""
            dropdown_elements = []
            for label, k in zip(label_list, key):
                name += label + k
                dropdown_elements.append(html.Span(label))  # Normal text
                dropdown_elements.append(html.Strong(k))    # Bold text

            dropdown_elements.append(html.Span(' ('))
            dropdown_elements.append(html.Strong(str(idx+1)))
            dropdown_elements.append(
                html.Span('/' + str(len(keys)) + ')'))  # Index text

            dropdown_label = html.Div(dropdown_elements, style={
                                      'whiteSpace': 'pre-wrap'})

            dropdown_value = repr(
                {
                    'index': idx,
                    'key': key,
                    'name': name
                }
            )
            options.append({'label': dropdown_label, 'value': dropdown_value})

        exp_name = exp_output.split('\'')[1]
        return options, options[0]['value'], html.Div(f"Experiment '{exp_name}' loaded.", id=ids.LOG)

    @app.callback(
        Output(ids.CURVE_DROPDOWN, 'value'),
        [Input(ids.BUTTON_BACK, 'n_clicks'),
         Input(ids.BUTTON_FORWARD, 'n_clicks')],
        [State(ids.CURVE_DROPDOWN, 'value'),
         State(ids.CURVE_DROPDOWN, 'options')],
        prevent_initial_call=True
    )
    def move_buttons(up_clicks, down_clicks, current_value, options):
        ctx = callback_context
        if not ctx.triggered:
            raise PreventUpdate
        trigger_id = ctx.triggered[0]['prop_id'].split('.')[0]

        current_index = next((i for i, option in enumerate(
            options) if option['value'] == current_value), None)

        if current_index is None:
            raise PreventUpdate

        if trigger_id == ids.BUTTON_FORWARD:
            new_index = min(len(options) - 1, current_index + 1)
        elif trigger_id == ids.BUTTON_BACK:
            new_index = max(0, current_index - 1)
        else:
            raise PreventUpdate

        if current_index == new_index:
            raise PreventUpdate

        return options[new_index]['value']

    return html.Div(
        children=[
            html.Button(
                children='<',
                id=ids.BUTTON_BACK,
                n_clicks=0,
                className="dash-button"
            ),
            dcc.Dropdown(
                id=ids.CURVE_DROPDOWN,
                options=[],
                clearable=False,
                style={
                    'width': '99%',
                    'justifyContent': 'space-around',
                    'margin': '0 auto',
                },
            ),
            html.Button(
                children='>',
                id=ids.BUTTON_FORWARD,
                n_clicks=0,
                className="dash-button"
            ),
        ],
        style={
            'display': 'flex',
            'width': '95%',
            'justifyContent': 'space-around',
            'margin': '0 auto',
        },
    )



import ast
import json
import pandas as pd
from io import BytesIO

from ...afm import AFM


def process_experiment(experiment: AFM, cp_data: str, vd_data: str, indentation: float | list[float]) -> tuple[AFM, pd.DataFrame]:
    cp_annotations = json.loads(cp_data)
    cp_annotations = {ast.literal_eval(
        key): value for key, value in cp_annotations.items()}

    vd_annotations = json.loads(vd_data)
    vd_annotations = {ast.literal_eval(
        key): value for key, value in vd_annotations.items()}

    experiment.experiment.update_annotations(cp_annotations)
    experiment.experiment.update_annotations(vd_annotations)

    df = experiment.experiment.get_fit_all(
        experiment.probe_diameter, ind=indentation)
    return experiment, df


def process_indentation(indentation: str) -> float | list[float]:
    indentation_list = indentation.split(';')
    ind = [float(i) for i in indentation_list]
    if len(ind) == 1:
        ind = ind[0]
    return ind


def get_pdf(experiment: AFM) -> BytesIO:
    pdf_merger = experiment.experiment.export_figures()

    pdf_bytes = BytesIO()
    pdf_merger.write(pdf_bytes)
    pdf_bytes.seek(0)

    return pdf_bytes



import pickle
import re
import io
import base64
from PIL import Image
import json
import os
from datetime import datetime


def get_current_annotation(current_dropdown_value, data: str):
    key = eval(current_dropdown_value)['key']
    annotations = json.loads(data)
    annotation = annotations[repr(key)]
    return annotation


def update_annotations(original_annotations: dict, loaded_annotations: dict) -> dict:
    for key, value in loaded_annotations.items():
        if key in original_annotations:
            original_annotations[key] = value
    return original_annotations


def make_folder(load_dir: str, name: str = "") -> str:
    if not name:
        time = datetime.now()
        name = time.strftime('%Y-%m-%d_%H-%M-%S') + '/'
    save_dir = load_dir + '/' + name
    os.mkdir(save_dir)
    return save_dir


def make_json(keys, value):
    dict = {}
    for key in keys:
        dict[repr(key)] = value
    return json.dumps(dict)


def load_json(content_string):
    decoded = base64.b64decode(content_string)
    loaded_json = json.loads(decoded.decode('utf-8'))
    return loaded_json


def load(encoded_data: str):
    decoded_data = bytes.fromhex(encoded_data)
    loaded_data = pickle.loads(decoded_data)
    return loaded_data


def dump(data):
    serialized_data = pickle.dumps(data)
    encoded_data = serialized_data.hex()
    return encoded_data


def extract_keys(name: str, labels: list[str]) -> tuple[str, ...]:
    regex_str = ""
    for l in labels:
        regex_str = regex_str + l + f'(?P<{l}>.*)'
    regex = re.compile(regex_str)
    matched = regex.match(name)
    if matched:
        key = tuple([matched.group(label) for label in labels])
        return key
    else:
        return ()


def group_values_by_keys(original_dict: dict) -> dict:
    if len(next(iter(original_dict))) == 1:
        return original_dict

    new_dict = {}
    for key, value in original_dict.items():
        new_key = key[:-1]
        if new_key not in new_dict:
            new_dict[new_key] = [value]
        else:
            new_dict[new_key].append(value)
    return new_dict

def load_image(image_path:str) -> str:
    img = Image.open(image_path)
    buffer = io.BytesIO()
    img.save(buffer, format="JPEG")  # can be changed to "PNG"
    encoded_image = base64.b64encode(buffer.getvalue()).decode()
    return f"data:image/jpeg;base64,{encoded_image}"